from textual.app import ComposeResult
from textual.widgets import Label, ListItem, ListView


class Sidebar(ListView):
    def compose(self) -> ComposeResult:
        yield ListItem(
            Label("App", classes="sidebar-item"), id="app-settings-container"
        )
        yield ListItem(
            Label("Typing", classes="sidebar-item"), id="typer-settings-container"
        )
        yield ListItem(
            Label("Content", classes="sidebar-item"), id="content-settings-container"
        )



import json
from typing import Any

from platformdirs import user_config_path

from k_hunter import const as CONST


class SettingsStorage:
    def __init__(self) -> None:
        self._config_path = (
            user_config_path(CONST.APP_NAME, ensure_exists=True)
            / CONST.SETTINGS_STORAGE_NAME
        )

    def load(self) -> dict[str, Any]:
        try:
            raw_json = self._config_path.resolve().read_text()
            return json.loads(raw_json)
        except FileNotFoundError:
            return {}

    def save(self, settings: dict[str, Any]) -> None:
        raw_json = json.dumps(settings)
        self._config_path.resolve().write_text(raw_json)



from typing import Sequence

from textual import events
from textual.app import ComposeResult
from textual.containers import (
    Center,
    CenterMiddle,
    Container,
    HorizontalGroup,
)
from textual.reactive import reactive
from textual.widgets import Label, Rule

from k_hunter.typer.schemas import Keystroke

from .schemas import TypingSessionSummary, TypingSummary
from .service import ProfileService


class StatItem(Container):
    def __init__(self, label: str, value: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self._label = label
        self._value = value

    def compose(self) -> ComposeResult:
        with Center():
            yield Label(content=self._label, classes="statistic-label")
        with Center():
            yield Label(content=self._value, classes="statistic-value")


class LastTypingSession(HorizontalGroup):
    typing_summary: reactive[TypingSessionSummary] = reactive(
        TypingSessionSummary(), recompose=True
    )
    BORDER_TITLE = "Last typing session"

    def compose(self) -> ComposeResult:
        yield StatItem("Accuracy", self.typing_summary.accuracy)
        yield StatItem("Speed", self.typing_summary.speed)


class TypingSessions(Container):
    typing_summary: reactive[TypingSummary] = reactive(TypingSummary(), recompose=True)

    def compose(self) -> ComposeResult:
        with Container(classes="stat-group width-auto height-auto horizontal"):
            yield StatItem("Time", self.typing_summary.time)
            yield StatItem("Typing sessions", self.typing_summary.typing_sessions)
        with Container(classes="stat-group width-auto height-auto horizontal"):
            yield StatItem("Top speed", self.typing_summary.speed_max)
            yield StatItem("Average speed", self.typing_summary.speed_avg)
        with Container(classes="stat-group width-auto height-auto horizontal"):
            yield StatItem("Top accuracy", self.typing_summary.accuracy_max)
            yield StatItem("Average accuracy", self.typing_summary.accuracy_avg)

    def on_resize(self, event: events.Resize):
        if event.container_size.width < 120:
            class_to_remove = "horizontal"
            class_to_add = "vertical"
        else:
            class_to_remove = "vertical"
            class_to_add = "horizontal"

        for container in self.query(".stat-group"):
            container.remove_class(class_to_remove)
            container.add_class(class_to_add)


class Profile(CenterMiddle):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._profile_service = ProfileService()

    def compose(self) -> ComposeResult:
        with Center():
            yield LastTypingSession(id="last_session")
        yield Rule(line_style="none")
        with Center():
            today_sessions = TypingSessions(id="today_sessions")
            today_sessions.border_title = "Statistics for today"
            yield today_sessions
        yield Rule(line_style="none")
        with Center():
            all_time_sessions = TypingSessions(id="all_time_sessions")
            all_time_sessions.border_title = "All time statistics"
            yield all_time_sessions

    def on_mount(self) -> None:
        self._update_statistic_widgets()

    def _update_statistic_widgets(self) -> None:
        self.query_one("#last_session", LastTypingSession).typing_summary = (
            self._profile_service.last_session
        )
        self.query_one("#today_sessions", TypingSessions).typing_summary = (
            self._profile_service.today
        )
        self.query_one("#all_time_sessions", TypingSessions).typing_summary = (
            self._profile_service.all_time
        )

    async def update_last_typing_result(
        self, typing_summary: Sequence[Keystroke]
    ) -> None:
        self._profile_service.add(typing_summary)
        self._update_statistic_widgets()



import sqlite3
from contextlib import contextmanager
from typing import Iterator

from platformdirs import user_data_path

from k_hunter import const as CONST


class SQLite3Storage:
    def __init__(self) -> None:
        self._db_path = (
            user_data_path(CONST.APP_NAME, ensure_exists=True)
            / CONST.TYPING_SESSIONS_STORAGE_NAME
        )

        self._create_tables()

    @contextmanager
    def _db_connection(self) -> Iterator[sqlite3.Connection]:
        conn = None
        try:
            conn = sqlite3.connect(
                database=self._db_path,
            )
            yield conn
            conn.commit()
        except Exception:
            if conn:
                conn.rollback()
            raise
        finally:
            if conn:
                conn.close()

    def add_session_summary(
        self, session_summary: dict, keystrokes_summary: list[dict]
    ) -> None:
        typing_session_stmt = """
        INSERT INTO
        typing_session (
          total_chars, correct_chars, speed, accuracy, elapsed_time_ms)
        VALUES (
          :total_chars, :correct_chars, :speed, :accuracy, :elapsed_time_ms)
        RETURNING id
        """
        keystrokes_summary_stmt = """
        INSERT INTO
        keystroke (
          typing_session_id, char, total, correct, accuracy, speed, elapsed_time_ms)
        VALUES (
          :typing_session_id, :char, :total, :correct, :accuracy, :speed, :elapsed_time_ms)
        """

        with self._db_connection() as conn:
            typing_session_id = conn.execute(
                typing_session_stmt, session_summary
            ).fetchone()

            for keystroke_summary in keystrokes_summary:
                keystroke_summary[CONST.TYPING_SESSION_ID_KEY] = typing_session_id[0]

            conn.executemany(keystrokes_summary_stmt, keystrokes_summary)

    def load_last_session_summary(self) -> tuple[int, int, int]:
        stmt = """
        SELECT
          speed, accuracy, elapsed_time_ms
        FROM typing_session
        ORDER BY id DESC LIMIT 1
        """

        with self._db_connection() as conn:
            return conn.execute(stmt).fetchone()

    def load_sessions_summary(
        self, today_only: bool
    ) -> tuple[int, int, int, int, int, int]:
        stmt = f"""
        SELECT
          count(*),
          cast(round(avg(speed)) as INTEGER),
          max(speed),
          cast(round(avg(accuracy)) as INTEGER),
          max(accuracy),
          sum(elapsed_time_ms)
        FROM typing_session
        {"WHERE DATE(date, 'localtime') = DATE('now', 'localtime')" if today_only else ""}
        """

        with self._db_connection() as conn:
            return conn.execute(stmt).fetchone()

    def _create_tables(self) -> None:
        stmt = """
        CREATE TABLE IF NOT EXISTS
        typing_session (
          id INTEGER PRIMARY KEY AUTOINCREMENT,
          total_chars INTEGER,
          correct_chars INTEGER,
          speed INTEGER,
          accuracy INTEGER,
          elapsed_time_ms INTEGER,
          date TIMESTAMP DEFAULT CURRENT_TIMESTAMP
        );

        CREATE INDEX IF NOT EXISTS idx_typing_session_date ON typing_session(date);

        CREATE TABLE IF NOT EXISTS
        keystroke (
          id INTEGER PRIMARY KEY AUTOINCREMENT,
          typing_session_id INTEGER,
          char TEXT,
          total INTEGER,
          correct INTEGER,
          accuracy INTEGER,
          speed INTEGER,
          elapsed_time_ms INTEGER,
          FOREIGN KEY (typing_session_id) REFERENCES typing_session(id) ON DELETE RESTRICT
        );

        CREATE INDEX IF NOT EXISTS idx_keystroke_char ON keystroke(char);
        """
        with self._db_connection() as conn:
            conn.executescript(stmt)



import random
from importlib import resources
from importlib.abc import Traversable

from k_hunter import const as CONST
from k_hunter.settings.schemas import ContentSettings

from .schemas import ContentType, NaturalLanguageCategory, ProgrammingLanguageCategory


class ContentService:
    def __init__(self, settings: ContentSettings) -> None:
        self.settings = settings
        self.content_files = resources.files(CONST.DATASETS_STORAGE_PATH)
        self.content_mapping = self._glob(self.content_files)

    def _glob(self, dir: Traversable) -> dict:

        rd = {}
        for sd in dir.iterdir():
            if sd.is_dir():
                rd[sd.name] = self._glob(sd)
            else:
                rd[sd.name] = sd

        return rd

    # Remove if not needed after release 1.0
    #
    # def content_types(self) -> list[str]:
    #     return [ct.capitalize().replace("_", " ") for ct in self.content_mapping.keys()]
    #
    # def languages(self, mode: str) -> list[str]:
    #     return list(self.content_mapping[mode].keys())
    #
    # def categories(self, mode: str, language: str) -> list[str]:
    #     return list(self.content_mapping[mode][language].keys())

    def category_files(
        self, content_type: str, language: str, category: str
    ) -> list[str]:
        return sorted(self.content_mapping[content_type][language][category].keys())

    def generate(self) -> str:
        match self.settings.content_type:
            case ContentType.NATURAL:
                return self._natural_language_text()
            case ContentType.PRAGRAMMING:
                return self._programming_language_text()

    def _natural_language_text(self) -> str:
        match self.settings.natural_language.category:
            case NaturalLanguageCategory.SIMPLE:
                return self._natural_language_simple_text()
            case NaturalLanguageCategory.COMMON:
                return self._natural_language_common_words()

    def _natural_language_simple_text(self) -> str:
        text = (
            self.content_files.joinpath(CONST.NATURAL_LANGUAGE_KEY)
            .joinpath(self._language_dir())
            .joinpath(CONST.SIMPLE_FILENAME)
            .read_text()
        )
        text = text.split("\n")
        return random.choice(text)

    def _natural_language_common_words(self) -> str:
        common_words = []
        content_files = self.settings.natural_language.common_words.content_files
        if not content_files:
            content_files = self.category_files(
                CONST.NATURAL_LANGUAGE_KEY,
                self.settings.natural_language.language.name.lower(),
                CONST.COMMON_DIR,
            )

        for filename in content_files:
            text = (
                self.content_files.joinpath(CONST.NATURAL_LANGUAGE_KEY)
                .joinpath(self._language_dir())
                .joinpath(CONST.COMMON_DIR)
                .joinpath(filename)
                .read_text()
            )
            common_words.extend(text.split())
        random.shuffle(common_words)

        words_count = self.settings.natural_language.common_words.words_count
        upper_percent = self.settings.natural_language.common_words.upper_percent

        words = [
            word.capitalize() if random.randint(0, 100) < upper_percent else word
            for word in common_words[:words_count]
        ]

        return "\n".join(words)

    def _programming_language_text(self) -> str:
        match self.settings.programming_language.category:
            case ProgrammingLanguageCategory.KEYWORDS:
                return self._programming_language_keywords()
            case ProgrammingLanguageCategory.CODE:
                return self._programming_language_code_samples()

    def _programming_language_keywords(self) -> str:
        keywords = []
        content_files = self.settings.programming_language.keywords.content_files
        if not content_files:
            content_files = self.category_files(
                CONST.PROGRAMMING_LANGUAGE_KEY,
                self.settings.programming_language.language.name.lower(),
                CONST.COMMON_DIR,
            )

        for filename in content_files:
            text = (
                self.content_files.joinpath(CONST.PROGRAMMING_LANGUAGE_KEY)
                .joinpath(self._language_dir())
                .joinpath(CONST.COMMON_DIR)
                .joinpath(filename)
                .read_text()
            )
            keywords.extend(text.split())

        keywords_count = self.settings.programming_language.keywords.keywords_count
        while len(keywords) < keywords_count:
            keywords.extend(keywords)

        random.shuffle(keywords)
        return "\n".join(random.choices(keywords, k=keywords_count))

    def _programming_language_code_samples(self) -> str:
        return "Available after adding CodeEngine"

    def _language_dir(self) -> str:
        match self.settings.content_type:
            case ContentType.NATURAL:
                settings = self.settings.natural_language
            case ContentType.PRAGRAMMING:
                settings = self.settings.programming_language
        return settings.language.name.lower()

    @property
    def placeholder(self) -> str:
        return "Press 'space' to start typing"



from typing import Final


#
# Settings
#
APP_NAME: Final = "k_hunter"
TYPING_SESSIONS_STORAGE_NAME: Final = "stat.db"
SETTINGS_STORAGE_NAME: Final = "settings.json"
DATASETS_STORAGE_PATH: Final = "k_hunter.content.datasets"
NATURAL_LANGUAGE_KEY: Final = "natural_language"
PROGRAMMING_LANGUAGE_KEY: Final = "programming_language"
SIMPLE_FILENAME: Final = "simple_text/simple"
COMMON_DIR: Final = "common"

THEME: Final = "nord"

SLE_WIDTH: Final = 70
SLE_MIN_WIDTH: Final = 50
SLE_MAX_WIDTH: Final = 120
SLE_WIDTH_STEPS: Final = 8
SLE_HEIGHT: Final = 1
SLE_MIN_HEIGHT: Final = 1
SLE_MAX_HEIGHT: Final = 1
SLE_START_FROM_CENTER: Final = True

SE_WIDTH: Final = 70
SE_MIN_WIDTH: Final = 50
SE_MAX_WIDTH: Final = 120
SE_WIDTH_STEPS: Final = 8
SE_HEIGHT: Final = 5
SE_MIN_HEIGHT: Final = 3
SE_MAX_HEIGHT: Final = 9
SE_HEIGHT_STEPS: Final = 4

WORDS_COUNT: Final = 50
MIN_WORDS_COUNT: Final = 20
MAX_WORDS_COUNT: Final = 100

UPPER_PERCENT: Final = 0
MIN_UPPER_PERCENT: Final = 0
MAX_UPPER_PERCENT: Final = 30

KEYWORDS_COUNT: Final = 50
MIN_KEYWORDS_COUNT: Final = 20
MAX_KEYWORDS_COUNT: Final = 100
#
# Field keys
#
THEME_KEY: Final = "theme"
LANGUAGE_KEY: Final = "language"
TYPER_KEY: Final = "typer"
CONTENT_KEY: Final = "content"

ENGINE_KEY: Final = "engine"
BORDER_KEY: Final = "border"
SLE_KEY: Final = "single_line_engine"
SE_KEY: Final = "standard_engine"

WIDTH_KEY: Final = "width"
MIN_WIDTH_KEY: Final = "min_width"
MAX_WIDTH_KEY: Final = "max_width"

HEIGHT_KEY: Final = "height"
MIN_HEIGHT_KEY: Final = "min_height"
MAX_HEIGHT_KEY: Final = "max_height"

SLE_START_FROM_CENTER_KEY: Final = "start_from_center"

CONTENT_TYPE_KEY: Final = "content_type"
CATEGORY_KEY: Final = "category"
CONTENT_FILES_KEY: Final = "content_files"
COMMON_WORDS_KEY: Final = "common_words"
WORDS_COUNT_KEY: Final = "words_count"
WORDS_MIN_COUNT_KEY: Final = "min_words_count"
WORDS_MAX_COUNT_KEY: Final = "max_words_count"
UPPER_PERCENT_KEY: Final = "upper_percent"
MIN_UPPER_PERCENT_KEY: Final = "min_upper_percent"
MAX_UPPER_PERCENT_KEY: Final = "max_upper_percent"
KEYWORDS_COUNT_KEY: Final = "keywords_count"
KEYWORDS_MIN_COUNT_KEY: Final = "min_keywords_count"
KEYWORDS_MAX_COUNT_KEY: Final = "max_keywords_count"
# CONTENT_LENGHT_KEY: Final = "content_lenght"  #
CODE_SAMPLE_TYPE_KEY: Final = "sample_type"

CHAR_KEY: Final = "char"
TOTAL_KEY: Final = "total"
CORRECT_KEY: Final = "correct"
TOTAL_CHARS_KEY: Final = "total_chars"
CORRECT_CHARS_KEY: Final = "correct_chars"
ELAPSED_TIME_MS_KEY: Final = "elapsed_time_ms"
ACCURACY_KEY: Final = "accuracy"
SPEED_KEY: Final = "speed"
TYPING_SESSION_ID_KEY: Final = "typing_session_id"

#
# Keystrokes
#
MILLISECONDS_MULTIPLIER: Final = 1000
PERCENT_SCALE_2DP: Final = 10_000
FLOAT_TO_INT_SCALE_2DP: Final = 100

#
# Typer
#
BORDER_EXPANSION: Final = 2



from typing import ClassVar

from textual.app import App, ComposeResult
from textual.binding import Binding, BindingType
from textual.widgets import ContentSwitcher, Footer

from k_hunter import const as CONST
from k_hunter.content.service import ContentService
from k_hunter.profile.service import ProfileService
from k_hunter.profile.widgets import Profile
from k_hunter.settings.messages import SettingChanged
from k_hunter.settings.schemas import AppSettings
from k_hunter.settings.service import SettingsService
from k_hunter.settings.widgets.settings_container import SettingsContainer
from k_hunter.typer.widgets import Typer, TyperContainer


class KeyHunter(App):
    CSS_PATH = "style.tcss"
    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("ctrl+t", "switch_widget('typer')", "Typing"),
        Binding("ctrl+s", "switch_widget('settings')", "Settings"),
        Binding("ctrl+p", "switch_widget('profile')", "Profile", priority=True),
    ]

    def __init__(self) -> None:
        super().__init__()
        self.profile_service = ProfileService()
        self.settings = AppSettings()
        self.settings_service = SettingsService(self.settings)
        self.content_service = ContentService(self.settings.content)

    def compose(self) -> ComposeResult:
        with ContentSwitcher(initial="typer"):
            yield TyperContainer(id="typer")
            yield Profile(id="profile")
            yield SettingsContainer(id="settings")

        yield Footer(show_command_palette=False)

    def action_switch_widget(self, widget_name: str) -> None:
        self.query_one(ContentSwitcher).current = widget_name

        self.screen.focus_next()

    def on_mount(self) -> None:
        self.watch(self.settings, CONST.THEME_KEY, self._on_theme_changed)

    def _on_theme_changed(self, theme: str) -> None:
        self.theme = theme

    def on_setting_changed(self, event: SettingChanged) -> None:
        self.settings_service.update(event.command)

    async def on_typer_typing_completed(self, event: Typer.TypingCompleted) -> None:
        await self.query_one(Profile).update_last_typing_result(event.typing_summary)
        self.action_switch_widget("profile")


def main():
    app = KeyHunter()
    app.run()


if __name__ == "__main__":
    main()



from typing import Any, Callable, ClassVar, Literal

from rich.style import Style
from textual import on
from textual.app import ComposeResult, RenderResult
from textual.binding import Binding, BindingType
from textual.color import Gradient
from textual.containers import HorizontalGroup
from textual.message import Message
from textual.renderables.bar import Bar as BarRenderable
from textual.validation import Validator
from textual.widgets import Input, Label, ProgressBar, Select, Switch
from textual.widgets._progress_bar import Bar
from textual.widgets._select import SelectOverlay

from k_hunter.settings.commands import SetSettingCommand
from k_hunter.settings.messages import SettingChanged
from k_hunter.settings.schemas import BaseSettings


class VimSelect(Select):
    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("enter,space,l", "show_overlay", "Show Overlay", show=False),
        Binding("up,k", "cursor_up", "Cursor Up", show=False),
        Binding("down,j", "cursor_down", "Cursor Down", show=False),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._type_to_search = False

    def action_cursor_up(self):
        if self.expanded:
            self.query_one(SelectOverlay).action_cursor_up()
        else:
            self.screen.focus_previous()

    def action_cursor_down(self):
        if self.expanded:
            self.query_one(SelectOverlay).action_cursor_down()
        else:
            self.screen.focus_next()


class SelectSetting(HorizontalGroup):
    def __init__(
        self,
        id: str,
        target: BaseSettings,
        attr_name: str,
        label: str,
        values: list[str],
        default: str | None = None,
        allow_blank: bool = False,
        cast: Callable[[Any], Any] | None = None,
    ) -> None:
        super().__init__(id=id, classes="setting-row")
        self.label = label
        self.values = values
        self.default = default
        self.allow_blank = allow_blank
        self._target = target
        self._attr_name = attr_name
        self._cast = cast

    def compose(self) -> ComposeResult:
        yield Label(self.label, classes="setting-label")
        with self.prevent(Select.Changed):
            yield VimSelect.from_values(
                values=self.values,
                allow_blank=self.allow_blank,
                value=self.default,
                classes="setting-select",
                compact=True,
            )

    def on_select_changed(self, event: Select.Changed) -> None:
        event.stop()

        self.post_message(
            SettingChanged(
                command=SetSettingCommand(
                    target=self._target,
                    attr_name=self._attr_name,
                    value=event.value,
                    cast=self._cast,
                ),
            )
        )


class InputSetting(HorizontalGroup):
    def __init__(
        self,
        id: str,
        target: BaseSettings,
        attr_name: str,
        label: str,
        default: str | int | float,
        validators: list[Validator],
        cast: Callable[[Any], Any] | None = None,
    ) -> None:
        super().__init__(id=id, classes="setting-row")
        self.label = label
        self.default = default
        self.validators = validators
        self._current_value = str(default)
        self._target = target
        self._attr_name = attr_name
        self._cast = cast

    def compose(self) -> ComposeResult:
        yield Label(content=self.label, classes="setting-label")
        with self.prevent(Input.Changed):
            yield Input(
                value=self._current_value,
                type=self._default_type(),
                validators=self.validators,
                select_on_focus=False,
                classes="setting-input",
                compact=True,
            )

    def _default_type(self) -> Literal["text", "integer", "number"]:
        match self.default:
            case str():
                return "text"
            case int():
                return "integer"
            case float():
                return "number"

    def _convert_to_default_type(self, value: str) -> str | int | float:
        match self.default:
            case str():
                return value
            case int():
                return int(value)
            case float():
                return float(value)

    def on_input_changed(self, message: Input.Changed) -> None:
        self.app.clear_notifications()
        if message.validation_result and not message.validation_result.is_valid:
            for failure in message.validation_result.failure_descriptions:
                self.notify(message=failure, title=self.label, severity="error")

    @on(Input.Submitted)
    @on(Input.Blurred)
    def process_input_submit(self, event: Input.Submitted | Input.Blurred) -> None:
        event.stop()
        new_value = event.value

        if new_value == self._current_value:
            return

        if event.validation_result and event.validation_result.is_valid:
            self.post_message(
                SettingChanged(
                    command=SetSettingCommand(
                        target=self._target,
                        attr_name=self._attr_name,
                        value=self._convert_to_default_type(new_value),
                        cast=self._cast,
                    ),
                )
            )
            self._current_value = new_value


class SwitchSetting(HorizontalGroup):
    def __init__(
        self,
        id: str,
        target: BaseSettings,
        attr_name: str,
        label: str,
        default: bool,
        cast: Callable[[Any], Any] | None = None,
    ) -> None:
        super().__init__(id=id, classes="setting-row")
        self.label = label
        self.default = default
        self._target = target
        self._attr_name = attr_name
        self._cast = cast

    def compose(self) -> ComposeResult:
        yield Label(content=self.label, classes="setting-label")
        yield Switch(
            value=self.default,
            classes="setting-switch height-auto",
        )

    def on_switch_changed(self, event: Switch.Changed) -> None:
        event.stop()
        self.post_message(
            SettingChanged(
                command=SetSettingCommand(
                    target=self._target,
                    attr_name=self._attr_name,
                    value=event.value,
                    cast=self._cast,
                ),
            )
        )


class ThumbStyle(BarRenderable):
    HALF_BAR_LEFT: str = "▐"
    BAR: str = "█"
    HALF_BAR_RIGHT: str = "▌"


class Thumb(Bar):
    def __init__(
        self,
        total_positions: float,
        *,
        id: str | None = None,
        classes: str | None = None,
        gradient: Gradient | None = None,
    ):
        super().__init__(
            id=id,
            classes=classes,
            gradient=gradient,
            bar_renderable=ThumbStyle,
        )
        self.thumb_ratio = 1 / total_positions

    def render(self) -> RenderResult:
        """Render the bar with the correct portion filled."""
        if self.percentage is None:
            return self.render_indeterminate()
        else:
            bar_style = self.get_component_rich_style("bar--bar")

            return self.bar_renderable(
                highlight_range=(
                    self.percentage * self.size.width,
                    (self.percentage + self.thumb_ratio) * self.size.width,
                ),
                highlight_style=Style.from_color(bar_style.color),
                background_style=Style.from_color(bar_style.bgcolor),
                gradient=self.gradient,
            )


class LinearSlider(ProgressBar, can_focus=True):
    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("minus,h,left", "decrease", "Decrease"),
        Binding("plus,l,right,equals_sign", "increase", "Increase"),
    ]

    class Changed(Message):
        def __init__(self, value: int) -> None:
            super().__init__()
            self.value: int = value

    def __init__(
        self,
        *,
        positions_count: int,
        current_value: int,
        min_value: int,
        max_value: int,
        id: str | None = None,
        classes: str | None = None,
        gradient: Gradient | None = None,
    ):
        super().__init__(
            positions_count,
            id=id,
            classes=classes,
            gradient=gradient,
        )
        if positions_count < 3:
            raise ValueError("Minimum three positions")

        self._positions_count = positions_count
        self._min_value = min_value
        self._max_value = max_value
        self._step = (self._max_value - self._min_value) / (self._positions_count - 1)
        self._current_position = self._compute_position(current_value)

    def compose(self) -> ComposeResult:
        yield Thumb(total_positions=self._positions_count, id="bar").data_bind(
            ProgressBar.percentage
        ).data_bind(ProgressBar.gradient)

    def on_mount(self) -> None:
        super().on_mount()
        self.update(advance=self._current_position)

    def action_increase(self) -> None:
        if self._current_position < self._positions_count - 1:
            self.update(advance=1)
            self._current_position += 1
            self.post_message(self.Changed(self._compute_value()))

    def action_decrease(self) -> None:
        if self._current_position > 0:
            self.update(advance=-1)
            self._current_position -= 1
            self.post_message(self.Changed(self._compute_value()))

    def set_value(self, value: int) -> None:
        if value == self._compute_value():
            return
        self._current_position = self._compute_position(value)
        self.update(
            progress=self._current_position,
        )

    def _compute_value(self) -> int:
        if self._current_position == 0 or self._positions_count <= 1:
            return self._min_value
        elif self._current_position == self._positions_count - 1:
            return self._max_value
        else:
            return self._min_value + int(round(self._current_position * self._step))

    def _compute_position(self, current_value: int) -> int:
        if current_value == self._min_value:
            return 0
        elif current_value == self._max_value:
            return self._positions_count - 1
        else:
            return int(round((current_value - self._min_value) / self._step))


class LinearSliderSetting(HorizontalGroup):
    def __init__(
        self,
        positions_count: int,
        current_value: int,
        min_value: int,
        max_value: int,
        id: str,
        label: str,
        target: BaseSettings,
        attr_name: str,
        cast: Callable[[Any], Any] | None = None,
    ) -> None:
        super().__init__(id=id, classes="setting-row")
        self.label = label
        self.positions_count = positions_count
        self.current_value = current_value
        self._min_value = min_value
        self._max_value = max_value
        self._target = target
        self._attr_name = attr_name
        self._cast = cast

    def compose(self) -> ComposeResult:
        yield Label(self.label, classes="setting-label")
        yield Label(
            f"{self.current_value:>3}",
            id="slider-value-label",
            classes="slider-value-label",
        )
        yield LinearSlider(
            positions_count=self.positions_count,
            current_value=self.current_value,
            min_value=self._min_value,
            max_value=self._max_value,
        )

    def on_linear_slider_changed(self, event: LinearSlider.Changed) -> None:
        self.query_one("#slider-value-label", Label).update(f"{event.value:>3}")
        self.post_message(
            SettingChanged(
                SetSettingCommand(
                    target=self._target,
                    attr_name=self._attr_name,
                    value=event.value,
                    cast=self._cast,
                )
            )
        )



from typing import Callable

from rich.segment import Segment
from rich.style import Style
from textual.strip import Strip

from k_hunter.settings.schemas import SingleLineEngineSettings

from .base_engine import BaseEngine


class SingleLineEngine(BaseEngine):
    def __init__(self, settings: SingleLineEngineSettings) -> None:
        super().__init__(settings)
        self._chars = []

        self._start_from_center = settings.start_from_center
        self._start_offset = (settings.width // 2) if settings.start_from_center else 0

    @property
    def start_from_center(self) -> bool:
        return self._start_from_center

    @start_from_center.setter
    def start_from_center(self, active: bool) -> None:
        self._start_from_center = active
        self.resize()

    @property
    def current_char(self) -> Segment:
        return self._chars[self._current_char_idx]

    @current_char.setter
    def current_char(self, current_char: Segment) -> None:
        self._chars[self._current_char_idx] = current_char

    @property
    def has_next(self) -> bool:
        return self._current_char_idx < (len(self._chars) - 1)

    def _update_current_char(self, style: Style) -> None:
        self.current_char = Segment(self.current_char.text, style)

    def _set_chars_style(self, get_char_style: Callable) -> None:
        self._chars = [
            Segment(
                text=char.text,
                style=get_char_style(self, char.style),
            )
            for char in self._chars
        ]

    def prepare_content(self, text: str) -> None:
        text = " ".join(text.split("\n"))

        before_chars = [
            Segment(" ", self.default_style) for _ in range(self._start_offset)
        ]

        self._chars = before_chars + [
            Segment(char, self.default_style) for char in text
        ]

        self._current_char_idx = self._start_offset
        self._update_current_char(self.next_char_style)

    def resize(self) -> None:
        if self.start_from_center:
            before_center = self._width // 2
        else:
            before_center = 0

        if self._chars:
            before_chars = [
                Segment(" ", self.default_style) for _ in range(before_center)
            ]

            self._chars = (
                before_chars + self._chars[self._start_offset : len(self._chars)]
            )

        self._current_char_idx = (
            self._current_char_idx - self._start_offset + before_center
        )
        self._start_offset = before_center

    def build_placeholder(self, y: int, text: str) -> Strip:
        if y != 0:
            return Strip.blank(self._width)

        text = f"{text:^{self._width}}"

        return Strip([Segment(char, self.default_style) for char in text])

    def build_line(self, y: int) -> Strip:
        if not self._chars or y != 0:
            return Strip.blank(self._width)

        if self.start_from_center:
            start = max(0, self._current_char_idx - self._start_offset)
        else:
            addition = self._width // 2
            if self._current_char_idx <= addition:
                start = 0
            else:
                start = max(0, self._current_char_idx - addition)
        end = min(start + self._width, len(self._chars))

        return Strip(self._chars[start:end])



from typing import Any

from textual.dom import DOMNode
from textual.reactive import reactive

from k_hunter import const as CONST
from k_hunter.content.schemas import (
    CodeSampleCategory,
    ContentType,
    NaturalLanguage,
    NaturalLanguageCategory,
    ProgrammingLanguage,
    ProgrammingLanguageCategory,
)
from k_hunter.typer.schemas import TyperBorder, TyperEngine

SettingsDict = dict[str, Any]


class BaseSettings(DOMNode):
    def __init__(self):
        super().__init__()
        for base in self.__class__.__bases__:
            self.__annotations__.update(base.__annotations__)

    def dump(self) -> SettingsDict:
        settings = {}
        for setting_name in self.__annotations__:
            setting = getattr(self, setting_name)
            if isinstance(setting, BaseSettings):
                settings[setting_name] = setting.dump()
            else:
                settings[setting_name] = setting

        return settings

    def load(self, settings: SettingsDict, set_reactive: bool = True) -> None:
        for setting_name in self.__annotations__:
            setting = getattr(self, setting_name)
            if isinstance(setting, BaseSettings):
                setting.load(settings.get(setting_name, {}), set_reactive=set_reactive)
            else:
                setting_type = type(setting)
                if setting_name not in self._reactives:
                    value = settings.get(setting_name)
                    if value is not None:
                        setattr(self, setting_name, value)

                    continue

                reactive_setting = self._reactives[setting_name]
                setting_default = reactive_setting._default
                value = setting_type(settings.get(setting_name, setting_default))
                if set_reactive:
                    self.set_reactive(reactive_setting, value)
                else:
                    setattr(self, setting_name, value)


class SizeConstraints(BaseSettings):
    width: reactive[int] = reactive(CONST.SLE_WIDTH, init=False)
    height: reactive[int] = reactive(CONST.SLE_HEIGHT, init=False)

    min_width = CONST.SLE_MIN_WIDTH
    max_width = CONST.SLE_MAX_WIDTH

    min_height = CONST.SLE_MIN_HEIGHT
    max_height = CONST.SLE_MAX_HEIGHT


class SingleLineEngineSettings(SizeConstraints):
    start_from_center: reactive[bool] = reactive(
        CONST.SLE_START_FROM_CENTER, init=False
    )


class StandardEngineSettings(SizeConstraints):
    height: reactive[int] = reactive(CONST.SE_HEIGHT, init=False)
    min_height = CONST.SE_MIN_HEIGHT
    max_height = CONST.SE_MAX_HEIGHT


class TyperSettings(BaseSettings):
    engine: reactive[TyperEngine] = reactive(TyperEngine.STANDARD, init=False)
    border: reactive[TyperBorder] = reactive(TyperBorder.HKEY, init=False)
    single_line_engine: SingleLineEngineSettings = SingleLineEngineSettings()
    standard_engine: StandardEngineSettings = StandardEngineSettings()


class NaturalLanguageCommonWordsSettings(BaseSettings):
    words_count: reactive[int] = reactive(CONST.WORDS_COUNT, init=False)
    min_words_count = CONST.MIN_WORDS_COUNT
    max_words_count = CONST.MAX_WORDS_COUNT

    upper_percent: reactive[int] = reactive(CONST.UPPER_PERCENT, init=False)
    min_upper_percent: reactive[int] = reactive(CONST.MIN_UPPER_PERCENT, init=False)
    max_upper_percent: reactive[int] = reactive(CONST.MAX_UPPER_PERCENT, init=False)

    content_files: list[str] = list()


class NaturalSimpleTextSettings(BaseSettings): ...


class NaturalLanguageSettings(BaseSettings):
    language: reactive[NaturalLanguage] = reactive(NaturalLanguage.EN, init=False)
    category: reactive[NaturalLanguageCategory] = reactive(
        NaturalLanguageCategory.COMMON, init=False
    )
    common_words: NaturalLanguageCommonWordsSettings = (
        NaturalLanguageCommonWordsSettings()
    )
    simple_text: NaturalSimpleTextSettings = NaturalSimpleTextSettings()


class KeywordsSettings(BaseSettings):
    keywords_count: reactive[int] = reactive(CONST.KEYWORDS_COUNT, init=False)
    min_keywords_count = CONST.MIN_KEYWORDS_COUNT
    max_keywords_count = CONST.MAX_KEYWORDS_COUNT
    content_files: list[str] = list()


class CodeSamplesSettings(BaseSettings):
    sample_type: reactive[CodeSampleCategory] = reactive(
        CodeSampleCategory.SIMPLE, init=False
    )


class ProgrammingLanguageSettings(BaseSettings):
    language: reactive = reactive(ProgrammingLanguage.PY, init=False)
    category: reactive[ProgrammingLanguageCategory] = reactive(
        ProgrammingLanguageCategory.KEYWORDS
    )
    keywords: KeywordsSettings = KeywordsSettings()
    code_samples: CodeSamplesSettings = CodeSamplesSettings()


class ContentSettings(BaseSettings):
    content_type: reactive[ContentType] = reactive(ContentType.NATURAL, init=False)
    natural_language: NaturalLanguageSettings = NaturalLanguageSettings()
    programming_language: ProgrammingLanguageSettings = ProgrammingLanguageSettings()


class AppSettings(BaseSettings):
    theme: reactive[str] = reactive(CONST.THEME, init=False)
    typer: TyperSettings = TyperSettings()
    content: ContentSettings = ContentSettings()



from typing import TYPE_CHECKING

from textual.app import ComposeResult
from textual.containers import Center, HorizontalGroup, VerticalGroup
from textual.widgets import Select, Switch

from k_hunter import const as CONST
from k_hunter.settings.schemas import TyperBorder
from k_hunter.typer.schemas import TyperEngine
from k_hunter.typer.simulator import TyperSimulator

from .components import (
    LinearSlider,
    LinearSliderSetting,
    SelectSetting,
    SwitchSetting,
)

if TYPE_CHECKING:
    from keyhunter.main import KeyHunter


class TyperEngineSelector(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        current_engine = self.app.settings.typer.engine
        available_engines = [engine.value for engine in TyperEngine]
        yield SelectSetting(
            id="typer_engine",
            label="Typer engine",
            values=available_engines,
            default=current_engine,
            target=self.app.settings.typer,
            attr_name=CONST.ENGINE_KEY,
            cast=TyperEngine,
        )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.typer,
            CONST.ENGINE_KEY,
            self._on_typer_engine_changed,
            init=False,
        )

    def _on_typer_engine_changed(self, engine: TyperEngine) -> None:
        with self.prevent(Select.Changed):
            self.query_one(Select).value = engine


class TyperBorderSelector(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        current_border = self.app.settings.typer.border
        available_borders = [border.value for border in TyperBorder]
        yield SelectSetting(
            id="typer_border",
            label="Border",
            values=available_borders,
            default=current_border,
            target=self.app.settings.typer,
            attr_name=CONST.BORDER_KEY,
            cast=TyperBorder,
        )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.typer,
            CONST.BORDER_KEY,
            self._on_typer_border_changed,
            init=False,
        )

    def _on_typer_border_changed(self, border: TyperBorder) -> None:
        with self.prevent(Select.Changed):
            self.query_one(Select).value = border


class SingleLineEngineWidth(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        settings = self.app.settings.typer.single_line_engine
        yield LinearSliderSetting(
            positions_count=CONST.SLE_WIDTH_STEPS,
            current_value=settings.width,
            min_value=settings.min_width,
            max_value=settings.max_width,
            id="sle-width",
            label="Width",
            target=self.app.settings.typer.single_line_engine,
            attr_name=CONST.WIDTH_KEY,
        )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.typer.single_line_engine,
            CONST.WIDTH_KEY,
            self._on_single_line_engine_width_changed,
            init=False,
        )

    def _on_single_line_engine_width_changed(self, width: int) -> None:
        with self.prevent(LinearSlider.Changed):
            self.query_one(LinearSlider).set_value(width)


class SingleLineEngineStartFromCenterSwitch(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        current_choice = self.app.settings.typer.single_line_engine.start_from_center
        yield SwitchSetting(
            id="sle_start_from_center",
            label="Start from center",
            default=current_choice,
            target=self.app.settings.typer.single_line_engine,
            attr_name=CONST.SLE_START_FROM_CENTER_KEY,
        )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.typer.single_line_engine,
            CONST.SLE_START_FROM_CENTER_KEY,
            self._on_sle_start_from_center_changed,
            init=False,
        )

    def _on_sle_start_from_center_changed(self, start_from_center: bool) -> None:
        with self.prevent(Switch.Changed):
            self.query_one(Switch).value = start_from_center


class StandardEngineWidth(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        settings = self.app.settings.typer.standard_engine
        yield LinearSliderSetting(
            positions_count=CONST.SE_WIDTH_STEPS,
            current_value=settings.width,
            min_value=settings.min_width,
            max_value=settings.max_width,
            id="se-width",
            label="Width",
            target=self.app.settings.typer.standard_engine,
            attr_name=CONST.WIDTH_KEY,
        )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.typer.standard_engine,
            CONST.WIDTH_KEY,
            self._on_standard_engine_width_changed,
            init=False,
        )

    def _on_standard_engine_width_changed(self, width: int) -> None:
        with self.prevent(LinearSlider.Changed):
            self.query_one(LinearSlider).set_value(width)


class StandardEngineHeight(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        settings = self.app.settings.typer.standard_engine
        yield LinearSliderSetting(
            positions_count=CONST.SE_HEIGHT_STEPS,
            current_value=settings.height,
            min_value=settings.min_height,
            max_value=settings.max_height,
            id="se-height",
            label="Height",
            target=self.app.settings.typer.standard_engine,
            attr_name=CONST.HEIGHT_KEY,
        )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.typer.standard_engine,
            CONST.HEIGHT_KEY,
            self._on_standard_engine_height_changed,
            init=False,
        )

    def _on_standard_engine_height_changed(self, height: int) -> None:
        with self.prevent(LinearSlider.Changed):
            self.query_one(LinearSlider).set_value(height)


class SingleLineEngineSettingsContainer(VerticalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        yield SingleLineEngineWidth(classes="setting-container")
        yield SingleLineEngineStartFromCenterSwitch(classes="setting-container")

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.typer, CONST.ENGINE_KEY, self._toggle_container_visibility
        )

    def _toggle_container_visibility(self, typer_engine: TyperEngine) -> None:
        if typer_engine == TyperEngine.SINGLE_LINE:
            self.remove_class("hidden")
        else:
            self.add_class("hidden")


class StandardEngineSettingsContainer(VerticalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        yield StandardEngineWidth(classes="setting-container")
        yield StandardEngineHeight(classes="setting-container")

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.typer, CONST.ENGINE_KEY, self._toggle_container_visibility
        )

    def _toggle_container_visibility(self, typer_engine: TyperEngine) -> None:
        if typer_engine == TyperEngine.STANDARD:
            self.remove_class("hidden")
        else:
            self.add_class("hidden")


class TyperSettingsContainer(VerticalGroup):
    app: "KeyHunter"

    BORDER_TITLE = "Typer"

    def compose(self) -> ComposeResult:
        yield TyperEngineSelector(classes="setting-container")
        yield TyperBorderSelector(classes="setting-container")

        yield SingleLineEngineSettingsContainer()
        yield StandardEngineSettingsContainer()

        self.typer = TyperSimulator(self.app.settings)
        self.typer.simulate()
        with Center():
            yield self.typer

    def on_show(self) -> None:
        self.typer.resume()

    def on_hide(self) -> None:
        self.typer.pause()



from typing import Callable

from rich.segment import Segment
from rich.style import Style
from textual.strip import Strip

from k_hunter.typer.base_engine import BaseEngine


class StandardEngine(BaseEngine):
    _lines = []

    @property
    def current_char(self) -> Segment | None:
        chars_count = 0
        for line in self._lines:
            if self._current_char_idx >= (len(line) + chars_count):
                chars_count += len(line)
            else:
                return line[self._current_char_idx - chars_count]

    @current_char.setter
    def current_char(self, current_char: Segment) -> None:
        chars_count = 0
        for line in self._lines:
            if self._current_char_idx >= (len(line) + chars_count):
                chars_count += len(line)
            else:
                line[self._current_char_idx - chars_count] = current_char
                return

    @property
    def has_next(self) -> bool:
        return self._current_char_idx < (sum([len(line) for line in self._lines]) - 1)

    @property
    def _current_line(self) -> int | None:
        chars_count = 0
        for line_index, line in enumerate(self._lines):
            if self._current_char_idx >= (len(line) + chars_count):
                chars_count += len(line)
            else:
                return line_index

    @property
    def _total_chars(self) -> int:
        return sum(len(line) for line in self._lines)

    def _update_current_char(self, style: Style) -> None:
        if current_char := self.current_char:
            self.current_char = Segment(current_char.text, style)

    def _set_chars_style(self, get_char_style: Callable) -> None:
        self._lines = [
            [Segment(char.text, get_char_style(self, char.style)) for char in line]
            for line in self._lines
        ]

    def _segmentize_word(self, word: str) -> list[Segment]:
        segments = [Segment(char, self.default_style) for char in word]
        segments.append(Segment(" ", self.default_style))

        return segments

    def _make_lines(self, words, f=lambda x: x) -> None:
        line = f(words[0])
        for word in words[1:]:
            if (len(line) + len(word)) < self._width:
                line.extend(f(word))
            else:
                self._lines.append(line)
                line = f(word)
        if line:
            self._lines.append(line)

    def prepare_content(self, text: str) -> None:
        self._lines.clear()

        words = text.split("\n")

        self._make_lines(words, self._segmentize_word)

        self._lines[-1].pop()

        self._current_char_idx = 0
        self._update_current_char(self.next_char_style)

    def resize(self) -> None:
        if not self._lines:
            return

        segmentized_text = []
        [segmentized_text.extend(line) for line in self._lines]

        self._lines.clear()

        words = []
        word_with_space = []
        for char in segmentized_text:
            word_with_space.append(char)
            if char.text == " ":
                words.append(word_with_space)
                word_with_space = []

        words.append(word_with_space)

        self._make_lines(words)

    def _blank_strip(self) -> Strip:
        return Strip([Segment(" ", self.default_style) for _ in range(self._width)])

    def build_placeholder(self, y: int, text: str) -> Strip:
        if y != self._height // 2:
            return self._blank_strip()

        text = f"{text:^{self._width}}"

        return Strip([Segment(char, self.default_style) for char in text])

    def build_line(self, y: int) -> Strip:
        if not self._lines or y >= self._height:
            return Strip.blank(self._width)

        middle = self._height // 2
        if self._current_line is not None and self._current_line > middle:
            y += self._current_line - middle

        if y > len(self._lines) - 1:
            return self._blank_strip()

        return Strip(self._lines[y])



from enum import Enum


class ContentType(str, Enum):
    NATURAL = "Natural language"
    PRAGRAMMING = "Programming language"


class NaturalLanguage(str, Enum):
    EN = "English"


class NaturalLanguageCategory(str, Enum):
    SIMPLE = "Simple text"
    COMMON = "Common words"


class ProgrammingLanguage(str, Enum):
    PY = "Python"


class ProgrammingLanguageCategory(str, Enum):
    KEYWORDS = "Keywords"
    CODE = "Code samples"


class CodeSampleCategory(str, Enum):
    SIMPLE = "Simple code"



from abc import ABC, abstractmethod
from typing import Callable

from rich.segment import Segment
from rich.style import Style
from textual.strip import Strip
from textual.theme import Theme


class BaseEngine(ABC):
    def __init__(self, settings) -> None:
        self._current_char_idx = 0

        self._width = settings.width
        self._min_width = settings.min_width
        self._max_width = settings.max_width

        self._height = settings.height
        self._min_height = settings.min_height
        self._max_height = settings.max_height

    @property
    def width(self) -> int:
        return self._width

    @width.setter
    def width(self, new_width: int) -> None:
        if new_width == self._width:
            return

        if new_width > self._max_width:
            self._width = self._max_width
        elif new_width < self._min_width:
            self._width = self._min_width
        else:
            self._width = new_width

        self.resize()

    @property
    def height(self) -> int:
        return self._height

    @height.setter
    def height(self, new_height: int) -> None:
        if new_height == self.height:
            return

        if new_height > self._max_height:
            self._height = self._max_height
        elif new_height < self._min_height:
            self._height = self._min_height
        else:
            self._height = new_height

        self.resize()

    @property
    @abstractmethod
    def current_char(self) -> Segment | None: ...

    @property
    def has_next(self) -> bool: ...

    @abstractmethod
    def _set_chars_style(self, get_char_style: Callable) -> None: ...

    @abstractmethod
    def _update_current_char(self, style: Style) -> None: ...

    def mark_current_char(self, is_matched: bool) -> None:
        if is_matched:
            self._update_current_char(self.matched_style)
        else:
            self._update_current_char(self.mismatched_style)

    def next(self) -> None:
        self._current_char_idx += 1
        self._update_current_char(self.next_char_style)

    def set_theme(self, theme: Theme) -> None:
        def get_char_style(self, style: Style | None) -> Style:
            match style:
                case self.matched_style:
                    return matched_style
                case self.mismatched_style:
                    return mismatched_style
                case self.next_char_style:
                    return next_char_style
                case _:
                    return default_style

        bgcolor = theme.background if theme.background else "#111111"
        default_style = Style(color=theme.foreground, bgcolor=bgcolor)
        matched_style = Style(color=theme.success, bgcolor=bgcolor)
        mismatched_style = Style(color=theme.error, bgcolor=bgcolor)
        next_char_style = default_style + Style(underline=True)

        self._set_chars_style(get_char_style)

        self.default_style = default_style
        self.matched_style = matched_style
        self.mismatched_style = mismatched_style
        self.next_char_style = next_char_style

    @abstractmethod
    def prepare_content(self, text: str) -> None: ...

    @abstractmethod
    def resize(self) -> None: ...

    @abstractmethod
    def build_placeholder(self, y: int, text: str) -> Strip: ...

    @abstractmethod
    def build_line(self, y: int) -> Strip: ...



import random

from k_hunter.typer.widgets import Typer


class TyperSimulator(Typer, can_focus=False):
    def simulate(self, pause: bool = True):
        self._test_content = self.app.content_service.generate()
        self._test_content_idx = 0
        self.engine.prepare_content(self._test_content)
        self._simulate_timer = self.set_interval(0.15, self._simulate_key, pause=pause)
        self._is_active_session = True

    def resume(self):
        self._simulate_timer.resume()

    def pause(self):
        self._simulate_timer.pause()

    def stop(self):
        self._simulate_timer.stop()

    def _simulate_key(self):
        if random.random() < 0.9:
            self.engine.mark_current_char(True)
        else:
            self.engine.mark_current_char(False)

        if self.engine.has_next:
            self.engine.next()
        else:
            self._simulate_timer.stop()
            self.simulate(pause=False)
        self.refresh()

    def _on_engine_changed(self) -> None:
        self.stop()
        super()._on_engine_changed()
        self.simulate(pause=False)



from typing import TYPE_CHECKING

from textual.app import ComposeResult
from textual.containers import HorizontalGroup, VerticalGroup
from textual.widgets import Select, SelectionList

from k_hunter import const as CONST
from k_hunter.content.schemas import (
    ContentType,
    NaturalLanguage,
    NaturalLanguageCategory,
    ProgrammingLanguage,
    ProgrammingLanguageCategory,
)
from k_hunter.settings.commands import SetSettingCommand
from k_hunter.settings.messages import SettingChanged

from .components import LinearSlider, LinearSliderSetting, SelectSetting

if TYPE_CHECKING:
    from k_hunter.main import KeyHunter


class ContentTypeSelector(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        content_type = self.app.settings.content.content_type.value
        available_content_types = [ct.value for ct in ContentType]
        with self.prevent(Select.Changed):
            yield SelectSetting(
                id="content-type",
                label="Content type",
                values=available_content_types,
                default=content_type,
                target=self.app.settings.content,
                attr_name=CONST.CONTENT_TYPE_KEY,
                cast=ContentType,
            )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content,
            CONST.CONTENT_TYPE_KEY,
            self._on_content_type_changed,
            init=False,
        )

    def _on_content_type_changed(self, content_type: ContentType) -> None:
        with self.prevent(Select.Changed):
            self.query_one(Select).value = content_type


class NaturalLanguageSelector(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        language = self.app.settings.content.natural_language.language.value
        available_languages = [lang.value for lang in NaturalLanguage]
        with self.prevent(Select.Changed):
            yield SelectSetting(
                id="natural-language",
                label="Language",
                values=available_languages,
                default=language,
                target=self.app.settings.content.natural_language,
                attr_name=CONST.LANGUAGE_KEY,
                cast=NaturalLanguage,
            )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content.natural_language,
            CONST.LANGUAGE_KEY,
            self._on_language_changed,
            init=False,
        )

    def _on_language_changed(self, language: NaturalLanguage) -> None:
        with self.prevent(Select.Changed):
            self.query_one(Select).value = language


class NaturalLanguageCategorySelector(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        category = self.app.settings.content.natural_language.category.value
        available_categories = [c.value for c in NaturalLanguageCategory]
        with self.prevent(Select.Changed):
            yield SelectSetting(
                id="natural-language-category",
                label="Category",
                values=available_categories,
                default=category,
                target=self.app.settings.content.natural_language,
                attr_name=CONST.CATEGORY_KEY,
                cast=NaturalLanguageCategory,
            )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content.natural_language,
            CONST.CATEGORY_KEY,
            self._on_category_changed,
            init=False,
        )

    def _on_category_changed(self, category: NaturalLanguageCategory) -> None:
        with self.prevent(Select.Changed):
            self.query_one(Select).value = category


class CommonWordsContent(HorizontalGroup):
    # TODO: Add watcher
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        content_files = self.app.content_service.category_files(
            CONST.NATURAL_LANGUAGE_KEY,
            self.app.settings.content.natural_language.language.name.lower(),
            CONST.COMMON_DIR,
        )
        selected = self.app.settings.content.natural_language.common_words.content_files
        selections = [
            (ct, ct, True if ct in selected else False) for ct in content_files
        ]
        sl = SelectionList(*selections)
        sl.border_title = "Words from"
        yield sl

    def on_selection_list_selected_changed(
        self, event: SelectionList.SelectedChanged
    ) -> None:
        if not event.control.selected:
            self.notify("Please select at least one source")
        else:
            self.post_message(
                SettingChanged(
                    command=SetSettingCommand(
                        target=self.app.settings.content.natural_language.common_words,
                        attr_name=CONST.CONTENT_FILES_KEY,
                        value=event.control.selected,
                    )
                )
            )


class CommonWordsCountContainer(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        settings = self.app.settings.content.natural_language.common_words
        yield LinearSliderSetting(
            positions_count=9,
            current_value=settings.words_count,
            min_value=settings.min_words_count,
            max_value=settings.max_words_count,
            id="common-words-count-setting",
            label="Words in session",
            target=self.app.settings.content.natural_language.common_words,
            attr_name=CONST.WORDS_COUNT_KEY,
        )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content.natural_language.common_words,
            CONST.WORDS_COUNT_KEY,
            self._on_common_words_count_changed,
            init=False,
        )

    def _on_common_words_count_changed(self, words_count: int) -> None:
        with self.prevent(LinearSlider.Changed):
            self.query_one(LinearSlider).set_value(words_count)


class UppercaseWordsPercentageContainer(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        settings = self.app.settings.content.natural_language.common_words
        yield LinearSliderSetting(
            positions_count=11,
            current_value=settings.upper_percent,
            min_value=settings.min_upper_percent,
            max_value=settings.max_upper_percent,
            id="common-words-upper-percent-setting",
            label="Percentage of uppercase words",
            target=self.app.settings.content.natural_language.common_words,
            attr_name=CONST.UPPER_PERCENT_KEY,
        )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content.natural_language.common_words,
            CONST.UPPER_PERCENT_KEY,
            self._on_common_words_upper_percent_changed,
            init=False,
        )

    def _on_common_words_upper_percent_changed(self, upper_percent: int) -> None:
        with self.prevent(LinearSlider.Changed):
            self.query_one(LinearSlider).set_value(upper_percent)


class ProgrammingLanguageSelector(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        language = self.app.settings.content.programming_language.language.value
        available_languages = [lang.value for lang in ProgrammingLanguage]
        with self.prevent(Select.Changed):
            yield SelectSetting(
                id="programming-language",
                label="Language",
                values=available_languages,
                default=language,
                target=self.app.settings.content.programming_language,
                attr_name=CONST.LANGUAGE_KEY,
                cast=ProgrammingLanguage,
            )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content.programming_language,
            CONST.LANGUAGE_KEY,
            self._on_language_changed,
            init=False,
        )

    def _on_language_changed(self, language: ProgrammingLanguage) -> None:
        with self.prevent(Select.Changed):
            self.query_one(Select).value = language


class ProgrammingLanguageCategorySelector(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        category = self.app.settings.content.programming_language.category.value
        available_categories = [c.value for c in ProgrammingLanguageCategory]
        with self.prevent(Select.Changed):
            yield SelectSetting(
                id="programming-language-category",
                label="Category",
                values=available_categories,
                default=category,
                target=self.app.settings.content.programming_language,
                attr_name=CONST.CATEGORY_KEY,
                cast=ProgrammingLanguageCategory,
            )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content.programming_language,
            CONST.CATEGORY_KEY,
            self._on_category_changed,
            init=False,
        )

    def on_setting_changed(self, event) -> None:
        event.stop()
        self.notify("Available after adding CodeEngine", severity="warning", timeout=4)
        self._on_category_changed(ProgrammingLanguageCategory.KEYWORDS)

    def _on_category_changed(self, category: ProgrammingLanguageCategory) -> None:
        with self.prevent(Select.Changed):
            self.query_one(Select).value = category


class CommonKeywordsContent(HorizontalGroup):
    # TODO: Add watcher
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        content_files = self.app.content_service.category_files(
            CONST.PROGRAMMING_LANGUAGE_KEY,
            self.app.settings.content.programming_language.language.name.lower(),
            CONST.COMMON_DIR,
        )
        selected = self.app.settings.content.programming_language.keywords.content_files
        selections = [
            (ct, ct, True if ct in selected else False) for ct in content_files
        ]
        sl = SelectionList(*selections)
        sl.border_title = "Words from"
        yield sl

    def on_selection_list_selected_changed(
        self, event: SelectionList.SelectedChanged
    ) -> None:
        if not event.control.selected:
            self.notify("Please select at least one source")
        else:
            self.post_message(
                SettingChanged(
                    command=SetSettingCommand(
                        target=self.app.settings.content.programming_language.keywords,
                        attr_name=CONST.CONTENT_FILES_KEY,
                        value=event.control.selected,
                    )
                )
            )


class KeywordsCountContainer(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        settings = self.app.settings.content.programming_language.keywords
        yield LinearSliderSetting(
            positions_count=9,
            current_value=settings.keywords_count,
            min_value=settings.min_keywords_count,
            max_value=settings.max_keywords_count,
            id="keywords-count-setting",
            label="Words in session",
            target=self.app.settings.content.programming_language.keywords,
            attr_name=CONST.KEYWORDS_COUNT_KEY,
        )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content.programming_language.keywords,
            CONST.KEYWORDS_COUNT_KEY,
            self._on_common_words_count_changed,
            init=False,
        )

    def _on_common_words_count_changed(self, keywords_count: int) -> None:
        with self.prevent(LinearSlider.Changed):
            self.query_one(LinearSlider).set_value(keywords_count)


class KeywordsSettingsContainer(VerticalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        yield KeywordsCountContainer(classes="setting-container")
        yield CommonKeywordsContent(classes="setting-container")

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content.programming_language,
            CONST.CATEGORY_KEY,
            self._toggle_container_visibility,
        )

    def _toggle_container_visibility(
        self, category: ProgrammingLanguageCategory
    ) -> None:
        if category == ProgrammingLanguageCategory.KEYWORDS:
            self.remove_class("hidden")
        else:
            self.add_class("hidden")


class CommonWordsSettingsContainer(VerticalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        yield CommonWordsCountContainer(classes="setting-container")
        yield UppercaseWordsPercentageContainer(classes="setting-container")
        yield CommonWordsContent(classes="setting-container")

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content.natural_language,
            CONST.CATEGORY_KEY,
            self._toggle_container_visibility,
        )

    def _toggle_container_visibility(self, category: NaturalLanguageCategory) -> None:
        if category == NaturalLanguageCategory.COMMON:
            self.remove_class("hidden")
        else:
            self.add_class("hidden")


class ProgrammingLanguageSettingsContainer(VerticalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        yield ProgrammingLanguageSelector(classes="setting-container")
        yield ProgrammingLanguageCategorySelector(classes="setting-container")
        yield KeywordsSettingsContainer()

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content,
            CONST.CONTENT_TYPE_KEY,
            self._toggle_container_visibility,
        )

    def _toggle_container_visibility(self, content_mode: ContentType) -> None:
        if content_mode == ContentType.PRAGRAMMING:
            self.remove_class("hidden")
        else:
            self.add_class("hidden")


class NaturalLanguageSettingsContainer(VerticalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        yield NaturalLanguageSelector(classes="setting-container")
        yield NaturalLanguageCategorySelector(classes="setting-container")
        yield CommonWordsSettingsContainer()

    def on_mount(self) -> None:
        self.watch(
            self.app.settings.content,
            CONST.CONTENT_TYPE_KEY,
            self._toggle_container_visibility,
        )

    def _toggle_container_visibility(self, content_mode: ContentType) -> None:
        if content_mode == ContentType.NATURAL:
            self.remove_class("hidden")
        else:
            self.add_class("hidden")


class ContentSettingsContainer(VerticalGroup):
    BORDER_TITLE = "Content"

    def compose(self) -> ComposeResult:
        yield ContentTypeSelector(classes="setting-container")
        yield NaturalLanguageSettingsContainer()
        yield ProgrammingLanguageSettingsContainer()



from typing import TYPE_CHECKING, ClassVar

from textual.app import ComposeResult
from textual.binding import Binding, BindingType
from textual.containers import VerticalScroll
from textual.widgets import ContentSwitcher, ListView

from .app_settings import AppSettingsContainer
from .content_settings import ContentSettingsContainer
from .settings_sidebar import Sidebar
from .typer_settings import TyperSettingsContainer

if TYPE_CHECKING:
    from k_hunter.main import KeyHunter


class SettingsContainer(VerticalScroll, can_focus=False):
    app: "KeyHunter"

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("ctrl+z", "undo", "Undo"),
        Binding("ctrl+r", "restore", "Restore"),
        Binding("ctrl+s", "save", "Save"),
        Binding("ctrl+d", "reset_to_default", "Default"),
    ]

    def compose(self) -> ComposeResult:
        yield Sidebar(id="sidebar")
        with ContentSwitcher(
            id="settings-group-switcher", initial="app-settings-container"
        ):
            yield AppSettingsContainer(
                id="app-settings-container", classes="settings-group-container"
            )
            yield TyperSettingsContainer(
                id="typer-settings-container", classes="settings-group-container"
            )
            yield ContentSettingsContainer(
                id="content-settings-container", classes="settings-group-container"
            )

    def action_undo(self) -> None:
        self.app.settings_service.undo()
        self.refresh_bindings()

    def action_restore(self) -> None:
        self.app.settings_service.restore()
        self.refresh_bindings()

    def action_save(self) -> None:
        self.app.settings_service.save()
        self.refresh_bindings()

    def action_reset_to_default(self) -> None:
        self.app.settings_service.reset_to_default()
        self.refresh_bindings()

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        if action in ("undo", "restore"):
            return True if self.app.settings_service.has_updates else None
        elif action == "save":
            return True if not self.app.settings_service.saved else None
        else:
            return True

    # Sidebar
    def on_list_view_highlighted(self, event: ListView.Highlighted) -> None:
        if event.item and event.item.id:
            self.query_one(ContentSwitcher).current = event.item.id

    def on_setting_changed(self) -> None:
        self.refresh_bindings()



from .commands import SetSettingCommand
from .schemas import AppSettings
from .storage import SettingsStorage


class SettingsService:
    def __init__(self, settings: AppSettings) -> None:
        self._history: list[SetSettingCommand] = []
        self._storage = SettingsStorage()
        self._settings = settings

        user_settings = self._storage.load()
        self._settings.load(user_settings)
        self._saved = True

    @property
    def has_updates(self) -> bool:
        for command in self._history:
            if command.executed:
                return True
        return False

    @property
    def saved(self) -> bool:
        return self._saved

    def update(self, command: SetSettingCommand):
        command.execute()
        self._history.append(command)
        self._saved = False

    def undo(self) -> None:
        for command in reversed(self._history):
            if command.executed:
                command.undo()
                self._saved = False
                break

    def restore(self) -> None:
        for command in reversed(self._history):
            command.undo()
            self._saved = False

    def reset_to_default(self) -> None:
        self.restore()
        self._settings.load({}, set_reactive=False)
        self._saved = False

    def save(self) -> None:
        self._storage.save(self._settings.dump())
        self._history.clear()
        self._saved = True



from typing import Any, Callable

from .schemas import BaseSettings


class SetSettingCommand:
    def __init__(
        self,
        target: BaseSettings,
        attr_name: str,
        value: Any,
        cast: Callable[[Any], Any] | None = None,
    ) -> None:
        self._target = target
        self._attr_name = attr_name
        self._value = value
        self._cast = cast
        self._executed = False

    @property
    def executed(self) -> bool:
        return self._executed

    def execute(self) -> None:
        if not self._executed:
            self._old_value = getattr(self._target, self._attr_name)
            value = self._cast(self._value) if self._cast else self._value
            setattr(self._target, self._attr_name, value)
            self._executed = True

    def undo(self) -> None:
        if self._executed:
            setattr(self._target, self._attr_name, self._old_value)
            self._executed = False



from typing import TYPE_CHECKING

from textual.app import ComposeResult
from textual.containers import HorizontalGroup, VerticalGroup
from textual.widgets import Select

from k_hunter import const as CONST

from .components import SelectSetting

if TYPE_CHECKING:
    from k_hunter.main import KeyHunter


class ThemeSelector(HorizontalGroup):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        current_theme = self.app.settings.theme
        available_themes = [
            theme_name
            for theme_name, theme in self.app.available_themes.items()
            if theme.dark and theme.name != "textual-ansi"
        ]
        with self.prevent(Select.Changed):
            yield SelectSetting(
                id="theme",
                label="Theme",
                values=available_themes,
                default=current_theme,
                target=self.app.settings,
                attr_name=CONST.THEME_KEY,
            )

    def on_mount(self) -> None:
        self.watch(
            self.app.settings, CONST.THEME_KEY, self._on_theme_changed, init=False
        )

    def _on_theme_changed(self, theme: str) -> None:
        with self.prevent(Select.Changed):
            self.query_one(Select).value = theme


class AppSettingsContainer(VerticalGroup):
    BORDER_TITLE = "App"

    def compose(self) -> ComposeResult:
        yield ThemeSelector(classes="setting-container")



from typing import NamedTuple


class TypingSessionSummary(NamedTuple):
    speed: str = "N/A"
    accuracy: str = "N/A"
    time: str = "00:00"


class TypingSummary(NamedTuple):
    time: str = "00:00:00"
    typing_sessions: str = "0"
    speed_avg: str = "N/A"
    speed_max: str = "N/A"
    accuracy_avg: str = "N/A"
    accuracy_max: str = "N/A"



from time import perf_counter
from typing import TYPE_CHECKING, Sequence

from textual import events
from textual.app import ComposeResult
from textual.containers import CenterMiddle
from textual.message import Message
from textual.strip import Strip
from textual.widget import Widget

from k_hunter import const as CONST
from k_hunter.settings.schemas import (
    AppSettings,
    TyperBorder,
    TyperEngine,
)

from .schemas import Keystroke
from .single_line_engine import SingleLineEngine
from .standard_engine import StandardEngine

if TYPE_CHECKING:
    from k_hunter.main import KeyHunter


class Typer(Widget, can_focus=True):
    app: "KeyHunter"

    class TypingCompleted(Message):
        def __init__(self, typing_summary: Sequence[Keystroke]) -> None:
            super().__init__()
            self.typing_summary = tuple(typing_summary)

    class TypingStarted(Message): ...

    def __init__(self, settings: AppSettings, **kwargs):
        super().__init__(**kwargs)

        self._is_active_session = False
        self._session_start_time_ms = 0
        self._keystroke_start_time_ms = 0
        self._keystrokes = []

        self._set_engine(settings)

    @property
    def _timer_ms(self) -> int:
        return round(perf_counter() * CONST.MILLISECONDS_MULTIPLIER)

    def _set_engine(self, settings: AppSettings) -> None:
        match settings.typer.engine:
            case TyperEngine.STANDARD:
                engine_settings = settings.typer.standard_engine
                self.engine = StandardEngine(engine_settings)
            case TyperEngine.SINGLE_LINE:
                engine_settings = settings.typer.single_line_engine
                self.engine = SingleLineEngine(engine_settings)

        self.engine.set_theme(self.app.available_themes[settings.theme])

        self.styles.height = engine_settings.height + CONST.BORDER_EXPANSION
        self.styles.width = engine_settings.width + CONST.BORDER_EXPANSION

    def on_mount(self) -> None:
        self._subscribe()

    def _subscribe(self) -> None:
        settings = self.app.settings
        self.watch(settings, CONST.THEME_KEY, self._on_theme_changed, init=False)
        self.watch(
            settings.typer, CONST.ENGINE_KEY, self._on_engine_changed, init=False
        )
        self.watch(settings.typer, CONST.BORDER_KEY, self._on_border_changed)

        sle_settings = settings.typer.single_line_engine
        self.watch(
            sle_settings, CONST.WIDTH_KEY, self._on_sle_width_changed, init=False
        )
        self.watch(
            sle_settings,
            CONST.SLE_START_FROM_CENTER_KEY,
            self._on_sle_start_from_center_changed,
            init=False,
        )

        se_settings = settings.typer.standard_engine
        self.watch(se_settings, CONST.WIDTH_KEY, self._on_se_width_changed, init=False)
        self.watch(
            se_settings, CONST.HEIGHT_KEY, self._on_se_height_changed, init=False
        )

    def _on_theme_changed(self, theme: str) -> None:
        self.engine.set_theme(self.app.available_themes[theme])

    def _on_border_changed(self, border: TyperBorder) -> None:
        self.styles.border = (border.value, self.styles.base.border_title_color)

    def _on_engine_changed(self) -> None:
        self._set_engine(self.app.settings)

    def _on_sle_start_from_center_changed(self, start_from_center: bool) -> None:
        if isinstance(self.engine, SingleLineEngine):
            self.engine.start_from_center = start_from_center

    def _on_sle_width_changed(self, width: int) -> None:
        if isinstance(self.engine, SingleLineEngine):
            self.engine.width = width
            self.styles.width = width + CONST.BORDER_EXPANSION

    def _on_se_width_changed(self, width: int) -> None:
        if isinstance(self.engine, StandardEngine):
            self.engine.width = width
            self.styles.width = width + CONST.BORDER_EXPANSION

    def _on_se_height_changed(self, height: int) -> None:
        if isinstance(self.engine, StandardEngine):
            self.engine.height = height
            self.styles.height = height + CONST.BORDER_EXPANSION

    def _process_keystroke(self, event: events.Key) -> None:
        if current_char := self.engine.current_char:
            current_time = self._timer_ms
            keystroke_elapsed_time_ms = current_time - self._keystroke_start_time_ms
            is_matched = current_char.text == event.character
            self._keystrokes.append(
                Keystroke(
                    key=current_char.text,
                    is_matched=is_matched,
                    elapsed_time_ms=keystroke_elapsed_time_ms,
                )
            )
            self._keystroke_start_time_ms = current_time
            self.engine.mark_current_char(is_matched)

            if self.engine.has_next:
                self.engine.next()
            else:
                self.stop_typing()

    def on_key(self, event: events.Key) -> None:
        if self._is_active_session:
            if event.key == "escape":
                self.stop_typing()
            else:
                self._process_keystroke(event)
        elif event.key == "space":
            self.engine.prepare_content(self.app.content_service.generate())
            self.start_typing()
        else:
            return None

        event.stop()
        event.prevent_default()
        self.refresh()

    def start_typing(self) -> None:
        self._keystrokes.clear()
        self._is_active_session = True
        self._session_start_time_ms = self._timer_ms
        self._keystroke_start_time_ms = self._session_start_time_ms
        self.post_message(self.TypingStarted())

    def stop_typing(self) -> None:
        self._is_active_session = False

        if self._keystrokes:
            self.post_message(self.TypingCompleted(typing_summary=self._keystrokes))

    def render_line(self, y: int) -> Strip:
        if not self._is_active_session:
            return self.engine.build_placeholder(
                y, self.app.content_service.placeholder
            )

        return self.engine.build_line(y)


class TyperContainer(CenterMiddle):
    app: "KeyHunter"

    def compose(self) -> ComposeResult:
        yield Typer(settings=self.app.settings)



from enum import Enum
from typing import NamedTuple


class TyperEngine(str, Enum):
    SINGLE_LINE = "Single line"
    STANDARD = "Standard"


class TyperBorder(str, Enum):
    BLANK = "blank"
    ROUND = "round"
    SOLID = "solid"
    THICK = "thick"
    DOUBLE = "double"
    HEAVY = "heavy"
    HKEY = "hkey"
    TALL = "tall"
    WIDE = "wide"


class Keystroke(NamedTuple):
    key: str
    is_matched: bool
    elapsed_time_ms: int



from datetime import datetime, timedelta
from typing import Sequence

from k_hunter import const as CONST
from k_hunter.typer.schemas import Keystroke

from .schemas import TypingSessionSummary, TypingSummary
from .storage import SQLite3Storage


class ProfileService:

    def __init__(self) -> None:
        self._storage = SQLite3Storage()
        self._last_session_summary = self._load_last_session_summary()
        self._today_summary = self._load_sessions_summary(today_only=True)
        self._all_time_summary = self._load_sessions_summary(today_only=False)

    @property
    def last_session(self) -> TypingSessionSummary:
        return self._last_session_summary

    @property
    def today(self) -> TypingSummary:
        return self._today_summary

    @property
    def all_time(self) -> TypingSummary:
        return self._all_time_summary

    def _load_last_session_summary(self) -> TypingSessionSummary:
        try:
            speed, accuracy, elapsed_time_ms = self._storage.load_last_session_summary()
            return TypingSessionSummary(
                speed=self._as_cpm(speed),
                accuracy=self._as_percent(accuracy),
                time=self._convert_time(elapsed_time_ms),
            )
        except TypeError:
            return TypingSessionSummary()

    def _load_sessions_summary(self, today_only: bool) -> TypingSummary:
        try:
            (
                sessions_count,
                speed_avg,
                speed_max,
                accuracy_avg,
                accuracy_max,
                sessions_time_ms,
            ) = self._storage.load_sessions_summary(today_only=today_only)

            return TypingSummary(
                time=self._convert_time(sessions_time_ms),
                typing_sessions=str(sessions_count),
                speed_avg=self._as_cpm(speed_avg),
                speed_max=self._as_cpm(speed_max),
                accuracy_avg=self._as_percent(accuracy_avg),
                accuracy_max=self._as_percent(accuracy_max),
            )
        except TypeError:
            return TypingSummary()

    def add(self, typing_summary: Sequence[Keystroke]) -> TypingSessionSummary:
        keystrokes = sorted(typing_summary, key=lambda x: x.key)
        keystrokes_summary = []
        session_summary = {
            CONST.TOTAL_CHARS_KEY: 0,
            CONST.CORRECT_CHARS_KEY: 0,
            CONST.ELAPSED_TIME_MS_KEY: 0,
        }

        last_char = keystrokes[0].key
        char_summary = {
            CONST.CHAR_KEY: last_char,
            CONST.TOTAL_KEY: 0,
            CONST.CORRECT_KEY: 0,
            CONST.ELAPSED_TIME_MS_KEY: 0,
        }

        for keystroke in keystrokes:
            if keystroke.key != last_char:
                keystrokes_summary.append(char_summary)
                last_char = keystroke.key
                char_summary = {
                    CONST.CHAR_KEY: last_char,
                    CONST.TOTAL_KEY: 0,
                    CONST.CORRECT_KEY: 0,
                    CONST.ELAPSED_TIME_MS_KEY: 0,
                }

            char_summary[CONST.TOTAL_KEY] += 1
            char_summary[CONST.CORRECT_KEY] += keystroke.is_matched
            char_summary[CONST.ELAPSED_TIME_MS_KEY] += keystroke.elapsed_time_ms
        else:
            if char_summary[CONST.TOTAL_KEY] > 0:
                keystrokes_summary.append(char_summary)

        for summary in keystrokes_summary:
            total_chars = summary[CONST.TOTAL_KEY]
            correct_chars = summary[CONST.CORRECT_KEY]
            elapsed_time_ms = summary[CONST.ELAPSED_TIME_MS_KEY]

            summary[CONST.ACCURACY_KEY] = self._compute_typing_accuracy(
                total_chars, correct_chars
            )
            summary[CONST.SPEED_KEY] = self._compute_typing_speed(
                total_chars, elapsed_time_ms
            )

            session_summary[CONST.TOTAL_CHARS_KEY] += total_chars
            session_summary[CONST.CORRECT_CHARS_KEY] += correct_chars
            session_summary[CONST.ELAPSED_TIME_MS_KEY] += elapsed_time_ms

        session_total_chars = session_summary[CONST.TOTAL_CHARS_KEY]
        session_correct_chars = session_summary[CONST.CORRECT_CHARS_KEY]
        session_elapsed_time_ms = session_summary[CONST.ELAPSED_TIME_MS_KEY]
        session_accuracy = self._compute_typing_accuracy(
            total_chars=session_total_chars,
            correct_chars=session_correct_chars,
        )
        session_speed = self._compute_typing_speed(
            total_chars=session_summary[CONST.TOTAL_CHARS_KEY],
            elapsed_time_ms=session_summary[CONST.ELAPSED_TIME_MS_KEY],
        )

        session_summary[CONST.ACCURACY_KEY] = session_accuracy
        session_summary[CONST.SPEED_KEY] = session_speed

        self._storage.add_session_summary(session_summary, keystrokes_summary)

        self._last_session_summary = TypingSessionSummary(
            speed=self._as_cpm(session_speed),
            accuracy=self._as_percent(session_accuracy),
            time=self._convert_time(session_elapsed_time_ms),
        )

        self._today_summary = self._load_sessions_summary(today_only=True)
        self._all_time_summary = self._load_sessions_summary(today_only=False)

        return self._last_session_summary

    def _convert_time(self, elapsed_time_ms: int) -> str:
        dt = datetime.min
        delta = timedelta(milliseconds=elapsed_time_ms)
        format = "%d days %H:%M:%S" if delta.days else "%H:%M:%S"
        return (dt + delta).strftime(format)

    def _as_cpm(self, speed: int) -> str:
        return f"{speed / CONST.FLOAT_TO_INT_SCALE_2DP}cpm"

    def _as_percent(self, accuracy: int) -> str:
        return f"{accuracy / CONST.FLOAT_TO_INT_SCALE_2DP}%"

    def _compute_typing_accuracy(self, total_chars: int, correct_chars: int) -> int:
        accuracy = round((correct_chars / total_chars) * CONST.PERCENT_SCALE_2DP)
        return int(accuracy)

    def _compute_typing_speed(self, total_chars: int, elapsed_time_ms: int) -> int:
        minutes = elapsed_time_ms / (60 * CONST.MILLISECONDS_MULTIPLIER)
        cpm = round((total_chars / minutes) * 100)
        return int(cpm)
